;********************************************************************
; SHOWMEM.ASM - display summary of all DOS memory blocks
;
; To assemble: ml showmem.asm hex.asm
;********************************************************************

.model small
.stack

	extern	hexfmt$bn:proc, hexfmt$wn:proc

CR			equ 	0dh
CRLF		textequ	<0dh, 0ah>
CRLFEnd 	textequ	<0dh, 0ah, '$'>
LF			equ 	0ah

;--------------------------------------------------------------------
; The structure of a Memory Control Block (MCB)

MCB struct 1
	marker		db	?				; M or Z (Z=last block in chain)
	owner_PSP	dw	?				; PSP seg addr of owning process
	size_para	dw	?				; size in paragraphs (16 bytes)
	reserved	db	3 dup( ? )
	progname	db	8 dup( ? )		; Name of owner (if DOS 4.0+)
MCB ends

;--------------------------------------------------------------------
.code
.startup

	; Print the program header

	mov ah, 9						; Clear the screen (print 50 LFs)
	mov dx, offset clear_screen
	int 21h
	mov ax, es						; Store (and print) the PSP addr
	mov [initial_PSP], ax
	mov di, offset PSP_address
	call hexfmt$wn
	mov dx, offset program_heading
	mov ah, 9
	int 21h

	; Print the MCB map on startup

	call show_memory_map
	call pause

	; Shrink the initial allocation block to the minimum

	mov bx, sp						; Get stack size (in bytes) and
	add bx, 15						; convert to paragraphs
	mov cl, 4
	shr bx, cl
	mov ax, ss						; Add stack size to stack seg
	add bx, ax
	mov ax, [initial_PSP]			; subtract the PSP seg addr
	mov es, ax
	sub bx, ax
	mov ah, 4ah						; shrink block to new size
	int 21h

	; Now shake things up a bit (we omit error checking on memory
	; allocations to make the code shorter & easier to understand.
	; You'll want to check if DOS sets the carry flag when you
	; allocate memory in your own programs!)

	mov es, ax						; Free the block holding the
	mov ax, es:[2ch]				; environment
	mov es, ax
	mov ah, 49h
	int 21h
	mov bx, 500H					; Get 500H paragraphs RIGHT AFTER
	mov ah, 48h 					; the PSP block
	int 21h
	mov bx, 02h 					; Set the allocation strategy to
	mov ax, 5801h					; "last fit"
	int 21h
	mov bx, 100H					; Get 100H paragraphs AT THE END
	mov ah, 48h
	int 21h
	mov bx, 200H					; Get 200H paragraphs AT THE END
	mov ah, 48h
	int 21h
	mov si, ax						; <store address for future use>
	mov bx, 400H					; Get 400H paragraphs AT THE END
	mov ah, 48h
	int 21h
	mov es, si						; Free the 200H paragraphs near
	mov ah, 49h 					; the end
	int 21h
	mov bx, 01h 					; Set the allocation strategy to
	mov ax, 5801h					; "best fit"
	int 21h
	mov bx, 150H					; Get 150H blocks
	mov ah, 48h
	int 21h
	mov es, di
	mov bx, 00h 					; Reset the allocation strategy
	mov ax, 5801h					; to "first fit"
	int 21h

	; Now show the new memory map after "shaking it up" a bit

	mov dx, offset clear_screen
	mov ah, 9
	int 21h
	call show_memory_map
	call pause

	mov ax, 4c00h					; Quit the program
	int 21h


;--------------------------------------------------------------------
;	pause : tell the user to "Press enter when ready" and then wait
;	for the user to press the enter key.
;
pause proc near

	mov dx, offset pause_msg		; Print the prompt
	mov ah, 9
	int 21h
@@:
	mov ah, 7						; Get a character
	int 21h
	cmp al, CR						; Did they press RETURN?
	jnz @B							; nope -- try again
	ret

pause endp


;--------------------------------------------------------------------
;	show_memory_map : Print the map of MCBs as created by DOS
;
show_memory_map proc near

	mov ah, 9
	mov dx, offset column_heading
	int 21h
	mov dx, offset column_underline
	int 21h

	; Find the DOS version

	mov si, 0						; SI = 0 if DOS ver < 4.0
	mov ax, 3000h
	int 21h
	cmp al, 4
	jl @F
	inc si
@@:


	; Get the segment address of the first memory block into ES

	mov ah, 52h 					; Put addr of LIST of LISTS
	int 21h 						; in ES:BX
	cmp bx, 0002h					; First MCB seg addr is at
	jge @F							; ES:[BX-2]. If BX<2, adjust
	mov ax, es						; ES & BX registers
	dec ax
	mov es, ax
	add bx, 10h
@@:
	mov ax, es:[bx-2]				; Get seg addr of First MCB
	mov es, ax						; and put it into ES

	; Print the current MCB data

print_mcb_block:
	mov ax, es						; the segment address
	mov di, offset output_line
	call hexfmt$wn
	mov al, ':' 					; the colon (SEG:OFF)
	mov [di], al
	inc di
	xor ax, ax						; the offset address
	call hexfmt$wn
	mov ax, '  '					; <two spaces>
	mov [di], ax
	inc di
	inc di
	mov ax, es:[MCB.owner_PSP]		; the owner PSP segment
	call hexfmt$wn
	mov ax, '  '					; <two spaces>
	mov [di], ax
	inc di
	inc di
	mov ax, es:[MCB.size_para]		; size of block in paragraphs
	call hexfmt$wn
	mov ax, '  '					; <two spaces>
	mov [di], ax
	inc di
	inc di
	mov bx, 0						; format the reserved area as a
@@: 								; series of hex bytes with spaces
	mov al, es:[bx+MCB.reserved]	; delimiting them
	call hexfmt$bn
	mov al, ' ' 					; <space>
	mov [di], al
	inc di
	inc bx
	cmp bx, 3
	jc @B
	mov [di], al
	inc di
	or si, si						; Is DOS version >= 4.0?
	jz end_of_line
	mov bx, es:[MCB.owner_PSP]		; Does this block hold the PSP?
	mov ax, es
	cmp ax, bx
	ja end_of_line
	add ax, es:[MCB.size_para]
	cmp ax, bx
	jna end_of_line
	mov bx, 0						; Print the program name
@@:
	mov al, es:[bx+MCB.progname]	; Get a character
	cmp al, ' ' 					; only print if char is printable
	jl end_of_line
	cmp al, '~'
	jg end_of_line
	mov [di], al
	inc di
	inc bx
	cmp bx, 8
	jle @B

end_of_line:
	mov ax, 0a0dh					; terminate the buffer
	mov [di], ax
	inc di
	inc di
	mov byte ptr [di],'$'
	mov ah, 9						; Print the MCB status line
	mov dx, offset output_line
	int 21h

	; Was this the last MCB block?

	mov al, es:[MCB.marker]
	cmp al, 'Z'
	jz done
	cmp al, 'M'
	jnz error

	; No it wasn't the last block ... go print the next

	mov ax, es						; The current MCB address
	add ax, es:[MCB.size_para]		; PLUS the size of the RAM area
	inc ax							; PLUS one gives address of the
	mov es, ax						; next MCB
	jmp print_mcb_block

	; We're done, so print the lines under the columns and quit

done:

	mov ah, 9
	mov dx, offset column_underline
	int 21h
	clc
	ret

	; Corrupted MCB chain! warn user and quit!

error:

	mov ah, 9
	mov dx, offset column_underline
	int 21h
	mov dx, offset corrupt_MCB_chain
	int 21h
	stc
	ret

show_memory_map endp

;--------------------------------------------------------------------
;	DATA AREA
.data

clear_screen		db	CRLF, 50 dup( LF ), '$'
column_heading		db	CRLF, CRLF, " Segment   ownr  Size  Bytes in"
					db	CRLF, " Address    PSP  para  rsvd fld    "
					db	"Name", CRLFEnd
column_underline	db	"---------  ----  ----  --------  --------"
					db	CRLFEnd
corrupt_MCB_chain	db	"ERROR-corrupt MCB chain!", CRLFEnd
program_heading 	db	"SHOWMEM - demonstrate how DOS manages RAM"
					db	CRLF, CRLF, "Current PSP address is: "
PSP_address			db  5 dup( '$' )
output_line 		db	100 dup( 0 )
pause_msg			db	"Press RETURN to continue:$"

initial_PSP 		dw	0

end
